// CPLEXConvert.c -- Kevin Leyton-Brown, October 2000
// This is a stand-alone file.

// The comments at the beginning refer to a slightly more expressive file
// format than the one generated by the programs in the test suite.
// However, it will work fine with the file format generated by the
// programs.

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>
#include <ctype.h>

// define CPLEXCONVERT as something else to avoid collision with CASS main
#define CPLEXCONVERT main

// allow different names for the stricmp function on non-windows builds
#ifndef _WIN32
#define stricmp strcasecmp
#endif

/* INPUT FILE FORMAT:
   ==================

% comments follow percentage symbols
% blank lines are ignored
% files are not case-sensitive
% [CR]'s are optional, all whitespace is ignored

goods 15
dummy 10
bids 800

% bidnum, price, goodnum, goodnum, ..., goodnum, #

0	20.75	2	6	8	#
1	30		1	2	5	7	13	#
2	5.20	4	#
  .
  .
  .

% implicitly, all goods numbered 15 or above (in this example) are dummies
% note that all counting starts from zero
% bidnum's do not need to be sequential between 0..799.  They just need to be unique.
*/

class CCInputBid
{
public:
	// data
	int *vec;
	unsigned amount, average;
	int bid_num;
	unsigned numGoods;
	
	// constructor/destructor
	CCInputBid(int a, int num);
	~CCInputBid();
	unsigned *array;
};

/* constructor */
CCInputBid::CCInputBid(int a, int num)
{
	vec = NULL;
	amount = a;
	bid_num = num;
	array = NULL;
}

/* destructor */
CCInputBid::~CCInputBid()
{
	if (vec) delete vec;
	if (array) delete[] array;
}


class CCInputFile
{
public:
	// this one is actually all you need
	bool parse (char *filename);

	// functions
	bool open (char *filename);
	void close();
	void parse();
	void print(int maxrec=-1);
	bool writeCPLEXFile(char *filename);

	// constructor/destructor
	CCInputFile();
	~CCInputFile();

	// variables
	unsigned int num_items, num_dummy_items, num_bids;
	CCInputBid **bid;

private:

	// functions
	void skipComments();
	void fileRead (unsigned int *int_input);
	void fileRead (float *float_input);
	void fileRead (const char *label, unsigned int *int_input);
	signed int fileReadGood (unsigned bidnum);
	bool isAllWhitespace(char *input,int len);
	
	// variables
	FILE *fp;
	char *filename, buffer[1000];
	int first_good;
	unsigned int *max_array, multi_unit;
};

/* constructor */
CCInputFile::CCInputFile()
{
	max_array = NULL;
	first_good = -1;
	multi_unit = 0;
	filename = NULL;
	bid = NULL;
}

/* destructor */
CCInputFile::~CCInputFile()
{
	if (filename) delete[] filename;
	if (bid)
	{
		for (unsigned int t=0; t<num_bids; t++)
			delete bid[t];
		delete[] bid;
	}

	if (max_array) delete[] max_array;
}

/* opens an input file, and leaves CCInputFile::fp pointing to the open file. */
bool CCInputFile::open(char *filename)
{
	this->filename = new char[strlen(filename)+1];
	strcpy(this->filename, filename);
	fp = fopen(filename, "rt");
	if (fp == NULL)
	{
		printf ("Error opening file \"%s\"\n",filename);
		return false;
	}
	return true;
}

/* closes an input file */
void CCInputFile::close()
{
	fclose(fp);
}

/* write a CPLEX input file */
// assumes that bid numbers are unique
bool CCInputFile::writeCPLEXFile(char *filename)
{
	FILE *fp2;
	unsigned i;

	fp2 = fopen(filename, "wt");
	if (fp == NULL)
	{
		printf ("Error writing CPLEX file \"%s\"\n",filename);
		return false;
	}

	// first section: maximize revenue
	fprintf (fp2, "max\n\n");
	for (i=0; i<num_bids; i++)
	{
		fprintf (fp2, "%dx%d ", bid[i]->amount, bid[i]->bid_num);
		if (i < num_bids - 1) fprintf (fp2, "+ ");
		if (i%10 == 9) fprintf(fp2,"\n");
	}

	// second section: constraints
	fprintf (fp2,"\n\nst\n\n");
	for (i=0; i<num_items; i++)
	{
		bool make_a_plus = false;
		int printed = 0;
		for (unsigned j=0; j<num_bids; j++)
		{
			// list each bid that names the current good
			if (bid[j]->vec[i])
			{
				if (make_a_plus) fprintf(fp2,"+ ");
				if(multi_unit)
					fprintf(fp2,"%dx%d ",bid[j]->vec[i],bid[j]->bid_num);
				else
					fprintf(fp2,"x%d ",bid[j]->bid_num);
				make_a_plus = true;
				if (printed++%10 == 9) fprintf(fp2,"\n");
			}
		}

		// finish off the constraint
		if(!make_a_plus) continue; // no variables for this constraint (no bids name this good)
		if(multi_unit)
			fprintf (fp2, " <= %d\n\n",max_array[i]);
		else
			fprintf (fp2, " <= 1\n\n");
	}

	// third section: integrality
	fprintf (fp2, "integer\n\n");
	for (i=0; i<num_bids; i++)
	{
		fprintf (fp2, "x%d ", bid[i]->bid_num);
		if (i%10 == 9) fprintf(fp2,"\n");
	}

	fprintf (fp2,"\n\nend\n");
	fclose(fp2);
	return true;
}

/* do everything in one command */
bool CCInputFile::parse (char *filename)
{
	if (!this->open (filename))
		return false;
	this->parse();
	this->close();
	return true;
}


/* parses the input file into a bunch of internal variables */
void CCInputFile::parse()
{
	unsigned int temp,t,tt;
	signed int foundGood;
	long pos;

	// read number of goods
	fileRead ("goods",&num_items);

	// skip dummy goods if necessary
	pos = ftell(fp);
	skipComments();
	fscanf(fp, "%s", buffer);
	if (!stricmp(buffer,"dummy"))
		// read number of dummy goods
		fileRead (&num_dummy_items);
	else
	{
		num_dummy_items = 0;
		fseek(fp,pos,SEEK_SET);
	}

	// allocate 'array' and 'bidcount'
	unsigned *array = new unsigned[num_items + num_dummy_items];

	// read number of bids
	fileRead ("bids", &num_bids);
	bid = new CCInputBid*[num_bids + num_items];  // + num_items so that we don't have to reallocate memory to add singleton bids later
	
	// is it a multi-unit file?
	pos = ftell(fp);
	skipComments();
	fscanf(fp, "%s", buffer);
	if (!stricmp(buffer,"maximums"))
		multi_unit=1;
	else
		multi_unit=0;
	fseek(fp,pos,SEEK_SET);

	if (multi_unit)
	{
		max_array = new unsigned[num_items];
		fileRead("maximums",&temp);
		max_array[0] = (int) temp;
		for (t=1;t<num_items;t++)
			fileRead(&(max_array[t]));
	}

	// read bids
	for (t=0; t<num_bids; t++)
	{
		// create a new bid
		bid[t] = new CCInputBid(0, 0);
		unsigned arrayindex = 0;
		
		// read bid number
		fileRead(&temp);
		bid[t]->bid_num = temp;
		
		// read bid price
		fileRead(&bid[t]->amount);
		
		// allocate and initialize goods array
		bid[t]->vec = new int[num_items + num_dummy_items];
		for (tt=0;tt<num_items + num_dummy_items;tt++) bid[t]->vec[tt] = 0;
		
		// read goods until end-of-record encountered
		foundGood = 0;
		for (tt=0;foundGood != -1;tt++)
		{
			foundGood = fileReadGood(bid[t]->bid_num);
			if (foundGood != -1)
			{
				if(multi_unit)
				{
					fileRead(&temp);
					bid[t]->vec[foundGood] = temp;
				}
				else
					bid[t]->vec[foundGood] = 1;
				array[arrayindex++] = foundGood;
			}
		}

		// copy the array to bid
		bid[t]->array = new unsigned[arrayindex];
		for (tt=0; tt<arrayindex; tt++)
			bid[t]->array[tt] = array[tt];
		bid[t]->numGoods = arrayindex;
	}

	delete[] array;
}

/* skip over comments */
void CCInputFile::skipComments()
{
	int pos;
	unsigned t;
	char c;

	do
	{
		pos = ftell(fp);
		fgets (buffer, 1000, fp);
		for (t=0,c=' ';t<strlen(buffer) && isspace(c);t++)
			c=buffer[t];  // allow comments to follow whitespace
	} while (!feof(fp) && (c == '%' || isAllWhitespace(buffer,strlen(buffer))));
	fseek (fp,pos,SEEK_SET);
}
	
bool CCInputFile::isAllWhitespace(char *input,int len)
{
	for (int t=0; t<len; t++)
		if (!isspace(input[t])) return false;
	return true;
}

/* read an integer from the file */
void CCInputFile::fileRead (unsigned int *int_input)
{
	skipComments();
	fscanf(fp, "%d", int_input);
}

/* read a float from the file */
void CCInputFile::fileRead (float *float_input)
{
	skipComments();
	fscanf(fp, "%f", float_input);
}

/* read a label followed by an integer from the file */
void CCInputFile::fileRead (const char *label, unsigned int *int_input)
{
	skipComments();
	fscanf(fp, "%s", buffer);
	if (stricmp(buffer,label))
	{
		printf("Error reading input file %s.\nExpected label %s, found %s.\n",filename, label, buffer);
		exit(1);
	}
	fscanf(fp, "%d", int_input);
}

/* read a good, unless '#' is the first thing found */
signed int CCInputFile::fileReadGood (unsigned bidnum)
{
	unsigned goodnum;
	
	skipComments();

	// read input
	fscanf(fp, "%s", buffer);

	// we found a pound sign -- end of record
	if (buffer[0] == '#')
		return -1;

	// we found a good number
	sscanf(buffer, "%d", &goodnum);
	if (goodnum >= num_items + num_dummy_items)
	{
		printf ("Good #%d was named in bid %d, but %d is the largest possible good #.\n",
			goodnum,bidnum,num_items + num_dummy_items);
		exit(1);
	}
	return goodnum;
}

/* output what has been read, after the read is complete */
void CCInputFile::print(int maxrec)
{
	printf ("Goods: %d, DGoods: %d, Bids: %d\n",num_items, num_dummy_items, num_bids);
	if (maxrec == -1) maxrec = num_bids;
	
	for (unsigned t=0; t<(unsigned) maxrec; t++)
		printf ("Bid#: %d, Amount: %d\n",bid[t]->bid_num, bid[t]->amount);
}


/* program starts here */
int CPLEXCONVERT(int argc, char *argv[], char *envp[]) 
{
	CCInputFile infile;
	char filename2[255], *filename;
	int counter;

	printf ("*******************************************************************************\n");
	printf ("* CPLEXConvert - CASS to CPLEX Conversion Utility                             *\n");
	printf ("* October 2000 Kevin Leyton-Brown, Stanford University                        *\n");
	printf ("* --------------------------------------------------------------------------- *\n");
	printf ("* This utility works for both single and multi-unit CASS files.               *\n");
	printf ("*                                                                             *\n");
	printf ("* Please send any feedback or bug reports to kevinlb@robotics.stanford.edu    *\n");
	printf ("*******************************************************************************\n");

	// parse parameters
	if (argc != 2)
	{
		printf ("Usage: CPLEXConvert file.txt\n\nThis will read in CASS file <file.txt> and write CPLEX file <file.lp>.\n");
		return 1;
	}

	/* read in the input file */
	filename = argv[1];
	if (!infile.parse(filename))
		exit (1);  // file could not be opened, error message has been displayed

	// CPLEX
	for (counter=0;filename[counter] != '.';counter++)
		filename2[counter] = filename[counter];
	filename2[counter++] = '.'; filename2[counter++] = 'l'; filename2[counter++] = 'p'; filename2[counter++] = 0;
	printf ("Writing CPLEX file %s...",filename2);
	infile.writeCPLEXFile(filename2);
	printf ("done.\n");

	return 0;
}

